;======================================================================
; SCIDYNE   649 School Street   Pembroke, MA 02359-3632 USA
; Tel: (781) 293-3059   Fax: (781) 293-4034   URL: www.scidyne.com
;======================================================================
;   --- DISCLAIMER ---  DISCLAIMER ---  DISCLAIMER --- DISCLAIMER ---
; This software may be used royalty-free, in whole or in part, only by
; original purchasers and owners of a SCIDYNE IOCP-74. The software is 
; supplied without any implied or expressed warranties or claims. The 
; user accepts all consequences resulting from its use and has the sole 
; responsibility in determining its suitability for each application.
;======================================================================
; Project:  IOCP-74  PC/104 Coprocessor utilizing a PIC16C74                                     
; File:     A7599A.ASM   IOCPbios                                               
; Revision: 1.01                                               
; Date:     12-08-98 
; Author:   Mark W. Durgin                                                  
;======================================================================
; This file comprises a basic operating system for the IOCP-74.  It is
; written in PIC assembly language and intended to be compiled using
; the macro assembler MPASM or MPLAB both available from Microchip
;======================================================================
; History:
;   12-08-98  R1.01
;   - Added software to extend RTS/TXEN beyond all bits of last character.
;		-	Changed timer-0 periodic cycle interrupts from 13.1ms to 1.64ms which
;			are now sub-cycle interrupts. Added sub-cycle counter to regain original
;		  13.1ms main loop execution rate.
;		- Added comm_aux_reg for additional serial communication functionality.
;		-	Added functions and ram locations to support basic digital input and
;     output operations.
;		-	All RAM locations (bank-0 and bank-1) are now cleared during bios initz.
;		  Removed unnecessary RAM initializations (individual clears).
;		-	Virtual stack is now setup earlier in initz so it can be correctly used
;			during initz routines.
;		- Added Parallel-Slave-Port write semaphore bits. New bytes written by the
;			host into the write array are now indentified by the corresponding
;			psp_wr_flag being set. Once a byte is acknowledged by the IOCP-74 application,
;			the flag must be manually reset. This procedure is unnecessary if words
;			(byte pairs) will not be written by the host.
;		-	Added two general purpose byte variables specifically for use during interrupts
;		-	Fixed CMD1 and CMD2 detection. Previoulsy failed detection if lower bits<5-0> had
;			any combination of set bits.
;
;   09-23-98  R0.00   - Beta release
;======================================================================

              list        p=16c74a      ;List directive to define processor
              errorlevel  -302          ;Suppress these messages
              radix       dec           ;Default radix will be decimal

#include <p16c74a.inc>                  ;Processor specific variable definitions

;  __CONFIG   _CP_OFF & _WDT_ON & _BODEN_ON & _PWRTE_ON & _RC_OSC

; '__CONFIG' directive is used to embed configuration data within .asm file.
; The lables following the directive are located in the respective .inc file.
; See respective data sheet for additional information on configuration word.

;--- Program constants and equates
FALSE                 equ 0         ;Logical FALSE condition
TRUE                  equ !FALSE    ;Logical TRUE condition
STACK_SIZE            equ 8         ;Size of virtual stack (in bytes)
COMM_BUF_SIZE         equ 16        ;Size of asynchronous communication buffers, RX and TX
PSP_ARY_SIZE          equ 16        ;Size of Parallel Slave Port RD & WR buffers (in bytes)
PSP_AUTO_INC_BIT      equ 5         ;Auto-Increment bit position
RTS_BIT               equ 0         ;Bit of PORTC used to for RTS and RS485 Driver enable
CR                    equ 0x0d      ;ASCII code for Carrige Return
DEFAULT_BAUD          equ 3         ;Defines default baud rate, 9600 
SUB_CYC_MODULUS       equ 8         ;Modulus of sub-cycle counter, determines main loop rate

;*****************************
; Add user equates here
;*****************************

;--- opsys_var0 bit definitions  
CYC_SYNC_BIT      equ 0     ;Cycle synchronization bit position
;                 equ 1     ;Reserved
;                 equ 2     ;Reserved
;                 equ 3     ;Reserved
;                 equ 4     ;Reserved
;                 equ 5     ;Reserved
;                 equ 6     ;Reserved
;                 equ 7     ;Reserved


;--- comm_ctrl_reg bit definitions
TX_BUSY_BIT       equ 0     ;TX control bit: 0=Not transmitting, 1=Busy transmitting
RX_ERROR_BIT      equ 1     ;RX error occured, see next three bits for reason
RX_OERR_BIT       equ 2     ;RX Overrun error occured
RX_FERR_BIT       equ 3     ;RX Framing error occured
RX_BOVR_BIT       equ 4     ;RX buffer over run
RX_TIMEOUT_BIT    equ 5     ;RX Timeout occured (reserved for future)
RX_MODE_BIT       equ 6     ;RX mode: (reserved for future) 0=REC til CR or buffer full, 1=REC til count or timeout
RX_ENABLE_BIT     equ 7     ;RX enable/done: 0=Disabled/done  1=Enabled/Busy receiving

;--- comm_aux_reg bit definitions
TX_LAST_CHAR_BIT  equ 0     ;TX last character bit: 0=Not last char, 1=Is last char
;                 equ 1     ;Reserved
;                 equ 2     ;Reserved
;                 equ 3     ;Reserved
;                 equ 4     ;Reserved
;                 equ 5     ;Reserved
;                 equ 6     ;Reserved
;                 equ 7     ;Reserved

;--- Define SPI selects
SPI0_USER0    equ 0     ;SPI select-0, User defined
SPI1_USER1    equ 1     ;SPI select-1, User defined
SPI2_USER2    equ 2     ;SPI select-2, User defined
SPI3_USER3    equ 3     ;SPI select-3, User defined
SPI4_ADC      equ 4     ;SPI select-4, Selects LTC1298 A/D
SPI5_DAC      equ 5     ;SPI select-5, Selects LTC1446 D/A
SPI6_EEMEM    equ 6     ;SPI select-6, Selects 25C16 EEPROM   
SPI7_INTREQ   equ 7     ;SPI select-7, Clocks flip-flop for host intr


;--- MACRO Definitions
 
;***********************************************
; This macro selects file registers in bank-0 
BANK0         macro
                bcf STATUS,RP0
              endm

;*********************************************** 
; This macro selects file registers in bank-1
BANK1         macro
                bsf STATUS,RP0
              endm

;***********************************************
; This macro pushes the contents of w on to
; the virtual stack.
PUSHW         macro
                call push_w_reg
              endm

;***********************************************
; This macro loads w from the most recent value
; pushed on the virtual stack.
POPW          macro
                call pop_w_reg
              endm

;*****************************************************************
; This macro performs one entire SPI transfer. Data being sent
; must already have been loaded into spi_outbuf. Data received
; during transfer is placed in spi_inbuf.
; Note: Assumes bank-0 is currently active
; SELECT    = SPI device which data will be communicated to/from
; BYTECOUNT = number of bytes to transfer
SPI_IO        macro select, bytecount
                movlw bytecount       ;Set number of bytes to xfer
                movwf spi_count       ;Preset counter
                movlw select          ; 
                call spi_select       ;Activate the SPI device
                call spi_xfer         ;Do the xfer
                SPI_DESELECT          ;Deselect the SPI device
              endm


;********************************************************************
; This macro sends and recieves one byte using SPI.  
; The PIC SPI hardware is controlled directly so that the normal
; SPI RAM buffers are not affected. The SPI device must already
; be selected and be deselected by other means.
; Note: Assumes bank-0 is currently active
SPI_1_BYTE    macro   
              local lbl0              ;Define locals of this macro 
                bcf PIR1,SSPIF        ;Clear the SSPIF
                movwf SSPBUF          ;Start byte transmision
lbl0:           btfss PIR1,SSPIF      ;Completed ?
                goto lbl0             ;No, wait
              endm


;**************************************************
; This macro deselects all SPI devices.  
; Note: Assumes bank-0 is currently active
SPI_DESELECT  macro  
                bcf SSPCON,4          ;Assure SPI clock is low when CS goes high 
                bsf PORTA,5           ;Disable the '138 decoder
                bsf SSPCON,4          ;Restore CKP = 1
              endm


;--- Program RAM Allocations 
                org 0x20  ;Start of RAM (Bank0)

  ;--- Temporary registers used for saving processor context
w_temp_0        RES 1     ;Temp register for W (interrupts) *** MUST be 1st variable
w_stack_temp0   RES 1     ;Temp register for stack usage    *** Must be 2nd
status_temp     RES 1     ;Temp register for STATUS
fsr_temp        RES 1     ;Temp register for FSR
pclath_temp     RES 1     ;Temp register for PCLATH

  ;--- System variables
opsys_var0      RES 1     ;Variable used by operating system
gp_byte_var0    RES 1     ;General purpose byte variable
intr_temp0			RES	1			;General purpose byte variable for interrupt routine use
intr_temp1			RES	1			;General purpose byte variable for interrupt routine use
timer0_sub_cyc  RES 1     ;Used to count sub-cycles of timer-0
  
  ;--- Math registers      *** Multiplcation ***      *** Division ***
math_word0       RES 2    ;Multiplier  (Result LW)    Dividend  (Quotiant)
math_word1       RES 2    ;Multiplicand               Divisor
math_word2       RES 2    ;Result HW                  Remainder
math_word3       RES 2    ;Used by routine            Used by routine
math_word4       RES 2    ;Used by routine            Not used

  ;--- SPI variables
spi_inbuf       RES 8     ;SPI data into the PIC 
spi_outbuf      RES 8     ;SPI data sent from the PIC
spi_count       RES 1     ;number of bytes to xfer
spi_temp0       RES 1     ;SPI temp register

  ;--- PSP variables
psp_write_ary       RES PSP_ARY_SIZE        ;This array is written by host and read by the IOCP-74   
psp_wr_flags        RES (PSP_ARY_SIZE / 8)  ;Semephore WR bits, each SET bit indicates a corresponding new byte
psp_read_ary        RES PSP_ARY_SIZE        ;This array is written by IOCP-74 and read by the host
psp_data_ary_pntr   RES 1                   ;PSP input/output array index pointer
psp_ctrl_reg        RES 1                   ;Control register for PSP operations

  ;--- Data acquisition specific variables
ao_ch0          RES 2     ;Holding register for LTC1458 ch-0
ao_ch1          RES 2     ;Holding register for LTC1458 ch-1
ai_ch0          RES 2     ;Holding register for LTC1298 ch-0
ai_ch1          RES 2     ;Holding register for LTC1298 ch-1 
dac_temp0       RES 1     ;Temp register for D/A calculations
dig_io_dir      RES 1     ;Digital I/O direction register
dig_io_data     RES 1     ;Digital I/O data register

                org 0xA0  ;Start of RAM (Bank1)
  
  ;--- These are mirror locations of variables in BANK-0 
  ;    which MUST be positioned here and remain in this order
                RES 1         ;Reserved and positioned to align with w_temp
                RES 1         ;Reserved and positioned to aling with w_stack_temp0

  ;--- Variables for virtual stack (Must reside in bank 1)
stack               RES STACK_SIZE  ;A virtual stack along with ...
stack_pntr          RES 1           ;   the stack pointer  
stack_status_temp   RES 1           ;   A register to save STATUS  
stack_fsr_temp      RES 1           ;   A register to save FSR  

  ;--- Variables for asyncronous communications
comm_ctrl_reg       RES 1               ;Communication control register  
comm_aux_reg        RES 1               ;Communication auxilary register
rx_comm_buf         RES COMM_BUF_SIZE   ;RAM buffer for incomming serial data
rx_comm_pntr        RES 1               ;Current position of rx buffer
;rx_timer            RES 1               ;Reserved for future receiver timeout timer  
tx_comm_buf         RES COMM_BUF_SIZE   ;RAM buffer for outgoing serial data
tx_comm_pntr        RES 1               ;Current position of tx buffer
tx_cntr             RES 1               ;Preset to # of outgoing bytes

;--- Processor specific vectors  
            org  0x000             ;Processor reset vector
            clrf PCLATH            ;Ensure page bits are cleared
            goto initz             ;Branch to start of program

            org  0x004             ;Interrupt vector location
intr_srv:   ;--- Save off processor context
            movwf   w_temp_0      ;Save off current W register contents
            movf    STATUS,w      ;Move status register into W register
            BANK0                 ;Ensure file register bank set to 0
            movwf   status_temp   ;Save off contents of STATUS register
            movf    FSR,w         ;Move FSR register into W register 
            movwf   fsr_temp      ;Save off contents of FSR register
            movf    PCLATH,w      ;Move PCLATH register into W register 
            movwf   pclath_temp   ;Save off contents of PCLATH register

            ;--- Test and service each interrupt source
            call psp_srv          ;Parallel Slave Port
            call cycle_srv        ;Periodic interrupts
            call comm_srv         ;Asynchronous communications
            
            ;**********************************
            ;*** Add user interrupt service 
            ;*** code or calls here
            ;**********************************

intr_srvx:  ;--- Restore processor context 
            BANK0                 ;Ensure file register bank set to 0
            movf    pclath_temp,w ;Retrieve copy of PCLATH register
            movwf   PCLATH        ;Restore pre-isr PCLATH register contents
            movf    fsr_temp,w    ;Retrieve copy of FSR register
            movwf   FSR           ;Restore pre-isr FSR register contents
            movf    status_temp,w ;Retrieve copy of STATUS register
            movwf   STATUS        ;Restore pre-isr STATUS register contents (and bank)
            swapf   w_temp_0,f    ;Restore pre-isr W register contents
            swapf   w_temp_0,w    ;  without affecting status bits
            retfie                ;Return from interrupt

;**************************************************************************
; MAIN -- This is the main loop and runs every 13.1ms. Timer-0 is free running
;   and will interrupt the PIC every 1.64ms. These interrupts (sub-cycles) are
;   counted by timer0_sub_cyc. After enough interrupts are reached to total
;   13.1ms the CYC_SYNC_BIT of opsys_var0 is set as a trigger to this main
;   loop.  When the main loop detects the CYC_SYNC_BIT set, it clears it 
;   and performs one pass through the application software. 
;   
;   There is nothing sacred about the sub-cycle interrupt period or the main
;   loop period being 1.64ms and 13.1ms. The timer-0 divider and SUB_CYC_MODULUS
;   can be changed to suit a user's particular application.

main:
      ;--- This loop is performed once every 13.1ms
loop:     BANK0                               ;Switch to bank-0
          btfss opsys_var0,CYC_SYNC_BIT       ;Test if time for action
          goto loop                           ;No, try again
          bcf opsys_var0, CYC_SYNC_BIT        ;Yes, clear bit for next time
          clrwdt                              ;Clear the watch-dog

		  ;***********************************
			; Add user application code here
			;***********************************

					goto loop


;         ==========================================================
;=======>>>>>> Start of IOCPbios functions and support routines <<<<<<============
;         ==========================================================

;*********************************************************************
;INITZ -- Initialize the IOCP-74 hardware and software
initz:                  
          ;--- Clear all RAM locations in BANK-0 and BANK-1
          BANK0
          movlw 0x20      ;Start of user bank-0 ram
          movwf FSR       ;Set indirect pointer
initz0:   clrf  INDF      ;Clear location in bank-0
					bsf	FSR,7				;Now point at into bank-1
					clrf INDF				;Clear location in bank-1
					bcf	FSR,7				;Now point back to bank-0
					incf  FSR,f     ;Inc the pointer
          btfss FSR,7     ;Were those the last locations?, bit 7 = 1?
          goto initz0  		;No, repeat

          ;--- Initz virtual stack
          BANK1                     ;Select register bank 1
          movlw stack+STACK_SIZE-1  ;Get address of virtual stack top
          movwf stack_pntr          ; and initz the stack pointer
          
          ;--- Setup PORTA and associated logic
																	;Bit: RA7  RA6  RA5  RA4  RA3  RA2  RA1  RA0 
                                  ;Fnc: Na   Na   SPEN A0   AN3  SPA2 AN1  AN0
          movlw b'00011011'       ;Dir: x    x    Out  In   In   Out  In   In 
          movwf TRISA             ;Set PORTA data direction register 
          BANK0                   ;Select bank 0
          bcf PORTA,2             ;Clr SPI addr 2 
          bsf PORTA,5             ;Deselect SPI select decoder
          BANK1                   ;Select bank 1
                    
          ;--- Setup PORTB and associated logic
                                  ;Bit: RB7  RB6  RB5  RB4  RB3  RB2  RB1  RB0
                                  ;Fnc: DIO7 DIO6 DIO5 DIO4 DIO3 SPA1 SPA0 DIO0
          movlw b'11111001'       ;Dir: In?  In?  In?  In?  In?  Out  Out  In?
          movwf TRISB             ;W -> portb data direction register
          BANK0                   ;Select bank 0
          bcf PORTB,1             ;Clr SPI addr 0
          bcf PORTB,2             ;Clr SPI addr 1
          BANK1                   ;Select bank 1
         
          ;--- Setup PORTC and associated logic
                                  ;Bit: RC7  RC6  RC5  RC4  RC3  RC2  RC1  RC0
                                  ;Fnc: Rxd  Txd  SDO  SDI  SCLK DIO2 DIO1 RTS
          movlw b'10010110'       ;Dir: IN   Out  Out  In   Out  In?  In?  Out
          movwf TRISC             ;W -> portc data direction register
          BANK0                   ;Select bank 0
          bsf PORTC,RTS_BIT       ;Deactivate RTS / Disable RS485 driver
          BANK1                   ;Select bank 1
          
          ;--- Setup PORTD and PORTE for Parallel Slave Port (PSP)
          movlw b'00010111'       ;Set RD, WR, and CS as
          movwf TRISE             ; inputs, Enable Parallel Slave Port
          movlw b'00000100'       ;Set RA0,RA1 and RA3 as analog inputs
          movwf ADCON1            ; and RA2,RA5 and RE<0-2> as digital
          bsf PIE1,PSPIE          ;Enable Parallel-Slave-Port interrupt 
          BANK0                   ;Switch to bank-0
          movf psp_read_ary+0,w   ;Get data in 1st read array location
          movwf PORTD             ;Store it for host access PORTD
                   
          ;--- Setup TIMER-0 for periodic interrupts
          BANK1
          bcf OPTION_REG,T0CS     ;Timer0 clock source = internal      
          bcf OPTION_REG,PSA      ;Prescaler assigned to timer0
          bcf OPTION_REG,PS0      ;Set prescaler value
          bcf OPTION_REG,PS1      ; 100 = Divide by 32
          bsf OPTION_REG,PS2      ; Cycle will occur every 1.64 Ms
          bsf INTCON,T0IE         ;Enable timer0 interrupt        

          BANK0
          movlw SUB_CYC_MODULUS   ;Get sub-cycle preload value
          movwf timer0_sub_cyc    ;Load the counter

          ;--- Setup TIMER-2 (Available to user or as SPI CLK) 
;          movlw 16                ;Set the modulus
;          movwf PR2
                                  
;          BANK0                   ;Select register page 0
;          movlw b'00000100'       ;Set pre-scaler of 1
;          movwf T2CON
;          BANK1                   ;Select register page 1
          
          ;--- Setup Synchronous-Serial-Port (SSP) for SPI operations
          BANK0                   ;Select register page 0
          movlw b'00110010'       ;Setup SPI as follows
          movwf SSPCON            ; Master mode, CLK = Fosc/64, CKP = 1 

          ;--- Setup for Asynchronous communications
          BANK1                   ;Select register page 1
          movlw DEFAULT_BAUD      ;Default baud rate is 9600 (Based on 20mhz clk)
          call setbaud            ;Set the baud rate
                                  
                                  ;-- RCSTA -- Receive status and control
          BANK0                   ;Bit:   7   6    5    4   3   2    1    0
          movlw b'10010000'       ;Fnc: SPEN RX9 SREN CREN Na FERR OERR RX9D
          movwf RCSTA             ; Enable serial communications and receiver                     
                     
                                  ;-- TXSTA -- Transmit status and control
          BANK1                   ;Bit:   7   6    5    4   3   2    1    0
          movlw b'00100000'       ;Fnc: CSRC TX9 TXEN SYNC Na BRGH TRMT TX9D
          movwf TXSTA             ; Enable the transmitter
                    
          bsf PIE1,RCIE           ;Enable async receive interrupts
           
          movlw rx_comm_buf       ;Set rx pointer 
          movwf rx_comm_pntr      ;  to beginning of buffer
           
          movlw tx_comm_buf       ;Set tx pointer to
          movwf tx_comm_pntr      ; beginning of buffer

          ;--- Initz the IOCP-74 digital I/O
          BANK0                   ;Switch to bank-0
          movlw 0xff        			;Have the IOCP digital port
          movwf dig_io_dir       	; be all inputs (dir all 1's)
          call  set_dig_dir       ;Setup the IOCP digital directions
           
          ;--- Enable interrupts
          bsf INTCON,PEIE         ;Enable peripheral interrupts
          bsf INTCON,GIE          ;Globally enable ALL interrupts

          goto main               ;Inialization done, Jump to main loop


;*********************************************************************  
; SET_DIG_DIR - Sets direction of IOCP digital I/O port    
;      
;   IOCP I/O    :  7   6   5   4   3   2   1   0
;   PIC I/O     : RB7 RB6 RB5 RB4 RB3 RC2 RC1 RB0
;
; Input:  Preload dig_io_dir with desired port directions where
;         0=output, 1 = input
;
; Output: None   
;
; Uses:   W, gp_byte_var0 

set_dig_dir:
              ;--- Set direction of I/O associated with PORTB
              BANK0                   ; Switch to bank-0
              movlw b'11111001'       ; Load W with Port B mask
              andwf dig_io_dir, w     ; SPI direction register left alone
              BANK1                   ; Switch to bank-1
              movwf TRISB             ; W -> Port B Data Direction Register 

              ;--- Set direction of I/O associated with PORTC
              movf  TRISC, w          ; TrisC -> W
              andlw b'11111001'       ; Clear Port C TRIS register Bits 1,2
              BANK0                   ; Switch to bank-0                                             
              movwf gp_byte_var0      ; Store result in Temp Register
              movf  dig_io_dir,w      ; Read the desired directions
              andlw b'00000110'       ; Extract only bits 1 & 2
              iorwf gp_byte_var0, w   ; Merge with existing TRISC data
              BANK1                   ; Switch to bank-1
              movwf TRISC             ; Update TRISC bits 1&2, Other bits unaffected
              return                  ; All Done


;*********************************************************************  
; UPDATE_DIG_IO - This routine updates the Digital I/O port and 
;                 dig_io_data register.    
;      
;   IOCP I/O    :  7   6   5   4   3   2   1   0
;   PIC I/O     : RB7 RB6 RB5 RB4 RB3 RC2 RC1 RB0
;
; Input   :   Pre-set dig_io_data with desired OUTPUT pattern. Bits 
;             corresponding to inputs are 'don't care'
;
; Output  :   Reads the IOCP-74 digital port and places the image into the    
;             dig_io_data register.
;
; Uses    :   W 

update_dig_io:
              ;--- Update digital outputs associated with PIC PORTB
                BANK0                   ; Switch to bank-0      
                movf  dig_io_data, w    ; Digital I/O Data -> W
                movwf PORTB             ; W -> Port B data Register
        
              ;--- Update digital outputs associated with PIC PORTC
                btfsc dig_io_data, 1    ; Should DIO.1 be 0 ? 
                goto  upd_dig0          ; No, branch
                bcf   PORTC, 1          ; Yes, clear DIO.1
                goto  upd_dig1          ; Do next bit
upd_dig0:       bsf   PORTC, 1          ; Set DIO.1

upd_dig1:       btfsc dig_io_data, 2    ; Should DIO.2 be 0 ?
                goto  upd_dig2          ; No, branch
                bcf   PORTC, 2          ; Yes, clear DIO.2
                goto  upd_dig3          ; Do next bit
upd_dig2:       bsf   PORTC, 2          ; Set DIO.2

              ;--- Update the digital inputs associated with PIC PORTB
upd_dig3:       movf  PORTB, w          ; Port B -> W
                andlw b'11111001'       ; Mask bits 1 and 2 of Port B
                movwf dig_io_data       ; Store results

              ;--- Update the digital inputs associated with PIC PORTC
                movf  PORTC, w          ; Port C -> W
                andlw b'00000110'       ; Mask bit 0 and 3 thru 7, Keep result in W
                iorwf dig_io_data, f    ; Combine Port B and C 
 
                return                  ; All Done


;*********************************************************************  
; SETBAUD - Sets baud rate for asynchronous communications  
;           Note: values based on 20mhz system clock    
;
; Input:  W = Baud rate code
;
; Output: Loads SPBRG   
;
; Uses:   W 

setbaud:  
            PUSHW                 ;Temp save w
            movlw HIGH baud_tbl   ;Get high byte of table address         
            movwf PCLATH          ;Make sure program counter is set properly
            POPW                  ;Restore w
         
            call baud_tbl         ;Look up the modulus
            BANK1                 ;Assure bank-1 is active
            movwf SPBRG           ;Set baud rate
            return                ;All done, return

            ;--- Table of setup codes to configure USART baud rate
baud_tbl:   addwf PCL,f   ;Add offset to pc, figure table entry
            retlw LOW 255     ;0: 1200 baud  (Actual = 1221, +1.73%)
            retlw LOW 129     ;1: 2400 baud  (Actual = 2404, +0.16%)
            retlw 64          ;2: 4800 baud  (Actual = 4808, +0.16%)
            retlw 32          ;3: 9600 baud  (Actual = 9469, -1.36%)
            retlw 15          ;4: 19200 baud (Actual = 19531, +1.73%)
            retlw 7           ;5: 38400 baud (Actual = 39063, +1.73%)


;*********************************************************************  
; SERIAL_RX - Sets up to serialy receive characters into RX buffers. 
;             Reception is done in background using interrupts 
;               
; Input:  None
;
; Output: None  
;
; Uses:   W, comm_ctrl_reg, rx_comm_buf, rx_comm_pntr 

serial_rx:
            BANK1                             ;Switch to bank-1
          
            ;--- Reset all RX error and condition reporting bits
            bcf comm_ctrl_reg,RX_ERROR_BIT       ;No errors of any type
            bcf comm_ctrl_reg,RX_OERR_BIT        ;No overrun errors 
            bcf comm_ctrl_reg,RX_FERR_BIT        ;No framing errors 
            bcf comm_ctrl_reg,RX_BOVR_BIT        ;No buffer overrun errors  
            bcf comm_ctrl_reg,RX_TIMEOUT_BIT     ;No RX timeout errors  

            ;--- Initz the RX buffer pointer
            movlw rx_comm_buf+0                   ;Set the pointer
            movwf rx_comm_pntr                    ; to start of buffer
          
            ;--- Enable receptions to RX buffer
            bsf comm_ctrl_reg,RX_ENABLE_BIT       ;Start a new reception to buffer
                     
            return                                ;All done


;*********************************************************************  
; SERIAL_TX - Starts a serial port transmission. Transmission is done
;             in background using interrupts  
;               
; Input:  - Preload comm_tx_buf with outgoing data
;         - Preload W with number of characters being sent
;
; Output: W returns False if TX not started or busy, Returns 1 if started 
;
; Uses:   W, comm_ctrl_reg, tx_comm_buf, tx_comm_pntr 

serial_tx:
            BANK1                             ;Switch to bank-1
            btfss comm_ctrl_reg,TX_BUSY_BIT   ;Already doing transmisions?
            goto serial_tx0                   ;No, proceed
            movlw FALSE                       ;Indicate TX not started or busy
            return                            ;Abort and try later
          
            ;--- Setup for interrupt driven serial async transmissions
serial_tx0: BANK0                             ;Switch to bank-0
            bcf PORTC,RTS_BIT                 ;Activate RTS / RS485 Drivers
            
            BANK1                             ;Switch to bank-1
            bsf   comm_ctrl_reg,TX_BUSY_BIT   ;Indicate transmitter now in use
            movwf tx_cntr                     ;Preload the TX count
            movlw tx_comm_buf+0               ;Get address of 1st outgoing byte     
            movwf tx_comm_pntr                ;Initz the pointer
           
            ;--- Software delay loop, allows time for RTS/TXEN to settle
            clrw                              ;Clear the W register         
serial_tx1: nop                               ;Waste some time ...
            nop                                
            nop                                
            nop                                  
            nop
            nop
            addlw 1                           ;Inc the loop counter
            btfss STATUS,Z                    ;Result zero?
            goto  serial_tx1                  ;No, do loop again
            
            bsf PIE1,TXIE                     ;Enable (start) transmit interrupts
           
            movlw TRUE                        ;Indicate to caller TX started
            return                            ;All done


;*********************************************************************  
; RD_EEMEM - Reads data from the serial EEPROM.  
;               
; Input:    W contains number of DATA BYTES to read. EEMEM starting 
;           location is placed in spi_outbuf.
;           as follows:
;             spi_outbuf+0      = EEMEM address LB
;             spi_outbuf+1      = EEMEM address HB
;             
; Output: Data read is placed in spi_inbuf. spi_inbuf(0) = 1st byte  
;
; Uses:   spi_temp1, W 

rd_eemem:
               
              PUSHW
               
              ;--- Swap the EEMEM address bytes because chip wants HB then LB 
              movf spi_outbuf+0,w     ;Get LB and
              PUSHW
              movf spi_outbuf+1,w     ;Get HB and
              movwf spi_outbuf+0      ; store it in 1st location
              POPW
              movwf spi_outbuf+1      ; store it in 2nd location

              movlw SPI6_EEMEM        ;Select EEMEM for SPI
              call spi_select         ; transfer

              movlw b'00000011'       ;EEMEM READ instruction
              SPI_1_BYTE

              movf spi_outbuf+0,w     ;EEMEM addr HB
              SPI_1_BYTE 

              movf spi_outbuf+1,w     ;EEMEM addr LB
              SPI_1_BYTE
              
              POPW
              movwf spi_count         ; and preload counter
              call spi_xfer           ;do the transfer
              SPI_DESELECT            ;deselct
            
              return                  ;All done


;*********************************************************************  
; WR_EEMEM -  Write data into the serial EEPROM. This routine assumes
;             the EEMEM is a 25C160 type device and write protection
;             on all block is disabled. See the 25C160 data sheet
;             for specific details concering its capabilities. 
;               
; Input:    W contains number of DATA BYTES to write. EEMEM starting 
;           location and data to be written is placed in spi_outbuf
;           as follows:
;             spi_outbuf+0      = EEMEM address LB
;             spi_outbuf+1      = EEMEM address HB
;             spi_outbuf+2..n   = Data to be stored in EEMEM
;
; Output:   none
;
; Uses:     W, spi_temp0, spi_temp1 

wr_eemem:
              addlw 2                 ;Inc the outgoing count by two
              PUSHW
              
              ;--- Swap the EEMEM address bytes because chip wants HB then LB 
              movf spi_outbuf+0,w     ;Get LB and
              PUSHW
              movf spi_outbuf+1,w     ;Get HB and
              movwf spi_outbuf+0      ; store it in 1st location
              POPW
              movwf spi_outbuf+1      ; store it in 2nd location

              ;--- Step one: Set the WREN bit in EEMEM chip
              movlw SPI6_EEMEM        ;Select EEMEM for SPI
              call spi_select         ; transfer
              
              movlw b'00000110'       ;EEMEM WREN instruction
              SPI_1_BYTE              ;Set the WREN bit, enable writes
              
              SPI_DESELECT            ;Bit actually gets set after CS goes high
              
              ;--- Step two: Send write instruction
              movlw SPI6_EEMEM        ;Again Select EEMEM for  
              call spi_select         ; SPI transfers
              
              movlw b'00000010'       ;EEMEM WRITE instruction
              SPI_1_BYTE              ;Send EEMEM WRITE instruction
              
              ;--- Step three: Send address and data to EEMEM
              POPW
              movwf spi_count         ; and preload counter
              call spi_xfer           ;Do the transfer
              SPI_DESELECT            ;Deselect the EEMEM
              
wr_ee0:       ;--- Step four: Wait for write to take       
              movlw b'00000101'       ;EEMEM RDSR instruction
              movwf spi_outbuf+0         
              SPI_IO  SPI6_EEMEM,2    ;Start byte transmision
              btfsc spi_inbuf+1,0     ;Is Write-In-Progress clear?
              goto wr_ee0             ;No, try again
              return                  ;All done
              
                
;*********************************************************************  
; RD_PIC_ADC - Reads a PIC ADC channel.  
;               
; Input:  Enter with W = 0, 1 or 2 corresponding to PIC ADC
;         channel RA0, RA1 or RA3 respectively. Entering with
;         W=3 shuts off the ADC. 
;
; Output: Returns 8-bit result in W  
;
; Uses:   W, gp_byte_var0 

rd_pic_adc:
            BANK0                   ;Switch to bank-0

            PUSHW                   ;Temp save w
            movlw HIGH pic_adc_tbl  ;Get high byte of table address         
            movwf PCLATH            ;Make sure program counter is set properly
            POPW                    ;Restore w
          
            call pic_adc_tbl        ;Retrive setup code for ADC
           
            movwf ADCON0            ;Load ADC with setup code and start conversion
            btfss ADCON0,ADON       ;Was ADC enabled?
            return                  ;No, return
             
            ;--- Wait required acquisition time
            movlw 65                ;Wait for acquisition ~13us
            movwf gp_byte_var0
pic_adc0:   decfsz gp_byte_var0,f
            goto pic_adc0

            bsf   ADCON0,GO_DONE    ;Start the conversion
pic_adc1:   btfsc ADCON0,GO_DONE    ;Conversion finished ?
            goto pic_adc1           ;No, try again

            movf ADRES,w            ;Read result
            return

            ;--- Array of setup codes to configure the PIC ADC
pic_adc_tbl:  addwf PCL,f           ;Add offset to pc, compute goto
              retlw b'10000001'     ;0: Select RA0, A/D Clk = Fosc/32 
              retlw b'10001001'     ;1: Select RA1, A/D Clk = Fosc/32 
              retlw b'10011001'     ;2: Select RA3, A/D Clk = Fosc/32 
              retlw b'10000000'     ;3: Turn off ADC


;*********************************************************************  
; ADC_UPDATE - Updates both channels of the LTC1298 Dual 12-bit ADC.  
;              Treats the channels as two independent single ended
;              inputs. 
; Input:  none    
;
; Output: ai_ch0 - updated with digitized analog voltage appearing at AI-0  
;         ai_ch1 - updated with digitized analog voltage appearing at AI-1
;
; Uses:   W, spi_inbuf, spi_outbuf 

adc_update:

          ;--- Read analog input channel 0
                              ;Bit: D7  D6  D5  D4  D3  D2  D1  D0
          movlw b'00000001'   ;Fnc:  0   0   0   0   0   0   0  STRT
          movwf spi_outbuf+0  ;Load 1st byte of out-going data  
 
                              ;Bit: D7  D6  D5  D4  D3  D2  D1  D0
          movlw b'10100000'   ;Fnc: S/D O/S M/L  x   x   x   x   x
          movwf spi_outbuf+1  ;Load 2nd byte of out-going data
                              ; Third byte is 'don't care'
                    
          SPI_IO SPI4_ADC,3   ;Do a SPI transfer of 3 bytes

          movf spi_inbuf+1,w  ;Read MSB data
          andlw 0x0f          ;Only lower four bits are valid
          movwf ai_ch0+1      ;Store new MSB of AI-CH0 
          movf spi_inbuf+2,w  ;Read LSB data
          movwf ai_ch0+0      ;Store new LSB of AI-CH0
          
          ;--- Read analog input channel 1
                              ;Bit: D7  D6  D5  D4  D3  D2  D1  D0
          movlw b'00000001'   ;Fnc:  0   0   0   0   0   0   0  STRT
          movwf spi_outbuf+0  ;Load 1st byte of out-going data  
 
                              ;Bit: D7  D6  D5  D4  D3  D2  D1  D0
          movlw b'11100000'   ;Fnc: S/D O/S M/L  x   x   x   x   x
          movwf spi_outbuf+1  ;Load 2nd byte of out-going data
                              ; Third byte is 'don't care'
                    
          SPI_IO SPI4_ADC,3   ;Do a SPI transfer of 3 bytes

          movf spi_inbuf+1,w  ;Read MSB data
          andlw 0x0f          ;Only lower four bits are valid
          movwf ai_ch1+1      ;Store new MSB of AI-CH1 
          movf spi_inbuf+2,w  ;Read LSB data
          movwf ai_ch1+0      ;Store new LSB of AI-CH1

          return              ;All done
            
                
;*********************************************************************  
; DAC_UPDATE - Updates both channels of the LTC1458 Dual 12-bit SPI5_DAC.  
;
; Input:  Preload ao_ch0 and ao_ch1 with 12-bit values to be written
;
; Output: none  
;
; Uses:   spi_outbuf, dac_temp0   

dac_update:

         ;--- Calc MS Byte
          swapf ao_ch0+0,w        ;Read and swap nibbles on SPI5_DACA LSB
          andlw 0x0f              ;Make sure bits 7-4 are 0
          movwf dac_temp0         ;Temp save it
          swapf ao_ch0+1,w        ;Read and swap nibbles on SPI5_DACA MSB
          andlw 0xf0              ;Make sure bits 3-0 are 0
          iorwf dac_temp0,w       ;Combine nibbles
          movwf spi_outbuf+0      ;Store ms byte

          ;--- Calc Mid byte
          swapf ao_ch0+0,w        ;Read and swap nibbles on SPI5_DACA LSB  
          andlw 0xf0              ;Make sure bits 3-0 are 0
          movwf dac_temp0         ;Temp save it
          movf  ao_ch1+1,w        ;Read and swap nibbles on SPI5_DACA LSB
          andlw 0x0f              ;Make sure bits 7-4 are 0
          iorwf dac_temp0,w       ;Combine nibbles
          movwf spi_outbuf+1      ;Store mid byte

          ;--- Calc LS byte
          movf  ao_ch1+0,w
          movwf spi_outbuf+2      ;Store ls byte 
           
          SPI_IO SPI5_DAC,3       ;Write three bytes to SPI5_DAC 
           
          return                  ;All done

          
;*****************************************************************************
; SPI_XFER - Uses PIC SSP module to do Serial Peripheral Interface
;   communications, shifting data in and out simultaneously. First byte
;   sent/recieved is located at spi_outbuf+0, spi_inbuf+0 respectivly. The SPI
;   device to be communicated to must already have been selected and be 
;   deselected by other means.
;   
; Note: If the SPI device being accessed requires a different clock polarity
;   or baud rate, these parameters must be configured before making this call. 
;
; Input:    spi_outbuf - Preload with any outgoing data
;           spi_count  - Preload with number of transfers to perform
;
; Output:   spi_inbuf  - Gets filled with any data shifted in
;
; Uses:     spi_outbuf, spi_inbuf. spi_temp0, spi_count  

spi_xfer:
     
          ;--- Perform SPI transfers
          clrf spi_temp0        ;Start pointer at 1st byte
          
spi_xfr1: movf spi_temp0,w      ;Read current index position
          addlw spi_outbuf      ;Calc offset into output buffer
          movwf FSR             ;Set indirect pointer  
          
          bcf PIR1,SSPIF        ;Clear the SSPIF
          movf INDF,w           ;Load a byte
          movwf SSPBUF          ;Start byte transmision
spi_xfr2: btfss PIR1,SSPIF      ;Completed ?
          goto spi_xfr2         ;No, wait

          movf spi_temp0,w      ;Read current index position
          addlw spi_inbuf       ;Calc offset into input buffer
          movwf FSR             ;Set indirect pointer

          movf SSPBUF,w         ;Read the data just shifted in
          movwf INDF            ;Store it in buffer

          incf spi_temp0,f      ;Inc the index
          decfsz spi_count,f    ;Dec the count, all done?
          goto spi_xfr1         ;No, do it again
          
          return


;*****************************************************************************
; SPI_SELECT - Selects (takes the *CS pin low) on the SPI device referenced
;   by the contents of W register. Manipulates bits on PORTA and PORTB
;   which are used as address inputs on a '138 3-8 decoder.
;
; Input:    W - SPI device to be enabled, 0-7
;
; Output:   None
;
; Uses:     spi_temp0   

spi_select:
              ;--- Set the desired select
              movwf spi_temp0       ;Save desired select for repeated examination
          
tstspi_a0:    btfsc spi_temp0,0     ;Test A0 bit
              goto set_spia0        ;Its 1, go and set A0
              bcf PORTB,1           ;Its 0, SPI A0 = 0
              goto tstspi_a1        ;Do next address bit
set_spia0:    bsf PORTB,1           ;SPI A0 = 1

tstspi_a1:    btfsc spi_temp0,1     ;Test A1 bit
              goto set_spia1        ;Its 1, go and set A1
              bcf PORTB,2           ;Its 0, SPI A1 = 0
              goto tstspi_a2        ;Do next address bit
set_spia1:    bsf PORTB,2           ;SPI A1 = 1

tstspi_a2:    btfsc spi_temp0,2     ;Test A2 bit
              goto set_spia2        ;Its 1, go and set A2
              bcf PORTA,2           ;Its 0, SPI A2 = 0
              goto spi_en           ;All done, enable device
set_spia2:    bsf PORTA,2           ;SPI A2 = 1

spi_en:       bcf PORTA,5           ;Enable device via 138 decoder

              return


;**************************************************************************
; GEN_HOST_IRQ - Generate a host interrupt
;   NOTE: Affects SPI decoder.. Don't call during a SPI xfer
; Input:    None
; Output:   None
; Uses:     None

gen_host_irq: bsf PORTB,1   ;SPI A0 = 1   
              bsf PORTB,2   ;SPI A1 = 1
              bsf PORTA,2   ;SPI A2 = 1
              bcf PORTA,5   ;Toggling decoder output clocks latch 
              bsf PORTA,5   ; and generates intr. Disable ALL SPI selects
              return


;*********************************************************************  
; PUSH_W_REG -  Contents of W is stored at current "stack" position and 
;               the "stack" pointer in decremented.
;
;     Note: This function is not re-entrent and therefore must not be 
;           called from an interrupt routine.

push_w_reg:          
            movwf w_stack_temp0           ;Temp save value being pushed
            movf  STATUS,w                ;Read the status register
            BANK1                         ;Ensure file register bank set to 1
            movwf stack_status_temp       ;Temp save the status register
            movf  FSR,w                   ;Get current FSR
            movwf stack_fsr_temp          ; and temp save it
            movf  stack_pntr,w            ;Get current "stack" position
            movwf FSR                     ;Set the indirect pointer
            btfss stack_status_temp,RP0   ;Was routine called while in bank 1?
            BANK0                         ;No switch to bank 0
            movf  w_stack_temp0,w         ;Recall the value to be pushed
            BANK1                         ;Ensure file register bank set to 1
            movwf INDF                    ;Store it in the virtual "stack"
            decf  stack_pntr,f            ;Dec the "stack" pointer for next push
            movf  stack_fsr_temp,w        ;Recall FSR
            movwf FSR                     ; and restore it
            movf  stack_status_temp,w     ;Recall pre-push status and
            movwf STATUS                  ; restore it (restores correct bank)
            movf  w_stack_temp0,w         ;Restore W
            return


;*********************************************************************  
; POP_W_REG - The most recent value pushed on the virtual "stack" 
;             is returned in W after the stack pointer in incremented 
;
;     Note: This function is not re-entrent and therefore can not be 
;           called from an interrupt routine

pop_w_reg:          
            movf  STATUS,w            ;Read the status register
            BANK1                     ;Switch to bank 1 
            movwf stack_status_temp   ; and temp save status
            movf  FSR,w               ;Read current FSR
            movwf stack_fsr_temp      ; and temp save it
            incf  stack_pntr,f        ;Inc the "stack" pointer
            movf  stack_pntr,w        ;Read the position
            movwf FSR                 ; and set the indirect pointer
            movf  INDF,w              ;Read value from "stack"
            movwf w_stack_temp0       ;Temp save it in bank 1
            BANK0                     ; and bank 0 since we aren't
            movwf w_stack_temp0       ; sure which bank is active
            BANK1                     ;Return to bank 1
            movf  stack_fsr_temp,w    ;Get saved FSR value
            movwf FSR                 ;restore it
            movf  stack_status_temp,w ;Recall pre-push status and
            movwf STATUS              ; restore it (restores correct bank)
            movf  w_stack_temp0,w     ;Get byte just "popped"
            return                    ; and return it in W


;****************************************************************
; MUL - Perform a 16-bit x 16-bit multiplcation resulting
;       in a 32-bit result.  
;                                                               
; Input:    math_word0 = 16-bit multiplier  
;           math_word1 = 16-bit multiplicand
;
; Output:   math_word0 = 32-bit result low word
;           math_word2 = 32-bit result high word                                                   
;
; Uses:     W, math_word<0-4>                                      

mul:
        movlw   16              ;Setup math_word4 as counter for for 16 shifts
        movwf   math_word4      

        movf    math_word0+1,W  ;Move math_word0 into math_word3
        movwf   math_word3+1
        movf    math_word0,W
        movwf   math_word3
        clrf    math_word0+1    ;Clr math_word0
        clrf    math_word0

mul0:   rrf     math_word3+1,F  ;Rotate math_word3 HB through carry
        rrf     math_word3,F    ;  and into LB
        btfss   STATUS,C        ;Test carry bit, skip if 0
        goto mul1

        ;--- Peform 16-bit addition
        movf    math_word1,W    ;Put math_word1 LB into W                       
        addwf   math_word0,F    ;Add W to math_word0 LB, result into math_word0  
        movf    math_word1+1,W  ;Put math_word1 HB into W                         
        btfsc   STATUS,C        ;Test carry, skip if clr                           
        incfsz  math_word1+1,W  ;Inc math_word1 HB, put in W, Skip if 0           
        addwf   math_word0+1,F  ;Add W to math_word0 HB, result in math_word0 HB   
        
mul1:   rrf     math_word0+1,F
        rrf     math_word0,F
        rrf     math_word2+1,F
        rrf     math_word2,F
        decfsz  math_word4,F
        goto    mul0
        movf    math_word0,W    ; Get low byte to W
        return


;********************************************************************
; DIV - Perform 16-bit division
;       ... math_word0 = math_word0 / math_word1
;
; Input:      math_word0 = Dividend
;             math_word1 = Divisor
;
; Output:     math_word0 = quotient, math_word2 = remainder
;
; Uses:       math_word<0-3>, w

div     movlw   16
        movwf   math_word3

        clrf    math_word2 + 1
        clrf    math_word2

div0    rlf     math_word0+1,W
        rlf     math_word2,F
        rlf     math_word2+1,F
        movf    math_word1,W
        subwf   math_word2,F
        movf    math_word1+1,W
        btfss   STATUS,C
        incfsz  math_word1+1,W
        subwf   math_word2+1,F

        btfsc   STATUS,C
        goto    div1
        movf    math_word1,W
        addwf   math_word2,F
        movf    math_word1+1,W
        btfsc   STATUS,C
        incfsz  math_word1+1,W
        addwf   math_word2+1,F

        bcf     STATUS,C

div1    rlf     math_word0,F
        rlf     math_word0+1,F

        decfsz  math_word3,F
        goto    div0
        movf    math_word0,W           
        return


;*******************************************************************
; ADD - Perform 16 bit addition, math_word0 = math_word0 + math_word1
; 
; Input   math_word0, math_word1 

; Output: Result in math_word0 
; 
; Uses:   math_word0, math_word1 

add     movf    math_word1,W    ;Put math_word1 LB into W
        addwf   math_word0,F    ;Add W to math_word0 LB, result into math_word0
        movf    math_word1+1,W  ;Put math_word1 HB into W
        btfsc   STATUS,C        ;Test carry, skip if clr
        incfsz  math_word1+1,W  ;Inc math_word1 HB, put in W, Skip if 0
        addwf   math_word0+1,F  ;Add W to math_word0 HB, result in math_word0 HB
        return



;********************************************************************** 
; COMM_SRV: Interrupt service routine for asynchronous communications 
;   Data is received into rx_comm_buf and transmited from tx_comm_buf

comm_srv:   

;--- Support for serial communication receive interrupts  
rx_srv:     ;--- Check for receive interrupts
            BANK0                 ;Ensure file register bank set to 0
            btfss PIR1,RCIF       ;Test USART receive interrupt flag
            goto tx_srv           ;No REC INTR, check for TRANS

            ;--- Check for receive framing error
            btfss RCSTA,FERR      ;Did framing error occur?
            goto chk_rx_oe        ;No, check for overrun error
            movf RCREG,w          ;Read and discard the rcv'd character, Clr RCIF & FERR
            
            ;--- Framing error detected
            BANK1                               ;Switch to bank-1
            btfss comm_ctrl_reg,RX_ENABLE_BIT   ;Was reception destined for the buffer?
            goto tx_srv                         ;No, no need to report, check for TRANS

            ;--- Report the framing error
            bsf comm_ctrl_reg,RX_FERR_BIT       ;Indicate framing error
            goto rx_error                       ;Finish RX error reporting


chk_rx_oe:  ;--- Check for receive overrun error          
            btfss RCSTA,OERR      ;Did overrun error occur?
            goto do_rx            ;No, ok to receive the byte
            
            ;--- Overrun error detected, clear the condition
            bcf RCSTA,CREN        ;Clear the CREN bit, also clears OERR
            bsf RCSTA,CREN        ;Re-enambe continuous receives
            movf RCREG,w          ;Read and discard the received character, Clr RCIF      
            
            ;--- Should error be reported?
            BANK1                               ;Switch to bank-1
            btfss comm_ctrl_reg,RX_ENABLE_BIT   ;Was reception destined for the buffer?
            goto tx_srv                         ;No, no need to report, check for TRANS

            ;--- Overrun error during reception to buffer
            bsf comm_ctrl_reg,RX_OERR_BIT       ;Indicate overrun error
            goto rx_error

do_rx:      ;--- No receive errors, proceed with normal reception
            BANK1                               ;Switch to bank-1
            btfsc comm_ctrl_reg,RX_ENABLE_BIT   ;Are receptions to buffer currently enabled?
            goto  norm_rx                       ;Yes, jump
            
            BANK0                 ;Switch to bank-0
            movf RCREG,w          ;Read and discard the received character, Clr RCIF      
            goto tx_srv           ;Done with REC INTR, check for TRANS
norm_rx:            
            movf  rx_comm_pntr,w  ;Get current buffer position
            movwf FSR             ;Set indirect pointer
            incf  rx_comm_pntr,f  ;Post-Inc and update the RX buffer pointer
            BANK0                 ;Switch to bank-0
            movf  RCREG,w         ;Read the received character, Clr RCIF
            movwf INDF            ;Store it in the buffer

            ;--- See if character just received was a Carrige Return
            sublw CR              ;Test for carrige return
            btfss STATUS,Z        ;Was it a CR?
            goto chk_rx_bf        ;No, check if buffer full

            ;--- Carrige return received, Reception done, cleanup
            BANK1                             ;Switch to bank-1
            bcf comm_ctrl_reg,RX_ENABLE_BIT   ;Disable receptions to RX buffer
            goto tx_srv                       ;Done with REC INTR, check for TRANS

chk_rx_bf:  ;--- Check if end of buffer was reached 
            BANK1                             ;Switch to bank-0
            movf  rx_comm_pntr,w              ;Is pointer now one byte beyond   
            sublw rx_comm_buf+COMM_BUF_SIZE   ; last valid RX buffer location?
            btfss  STATUS,Z                   ;      (same value?)
            goto tx_srv                       ;No, continue on to TRANS intr
            bsf comm_ctrl_reg,RX_BOVR_BIT     ;Yes, Indicate RX buffer overrun error

rx_error:   ;--- A reception error occured. Finish reporting the error        
            bcf comm_ctrl_reg,RX_ENABLE_BIT     ;Indicate done and disable buffer receptions            
            bsf comm_ctrl_reg,RX_ERROR_BIT      ;Indicate an error occured
            goto tx_srv                         ;Done with REC INTR, check for TRANS

;--- Support for serial communication transmit interrupts  
tx_srv:     ;--- Check for transmit interrupts
            BANK1
            btfss PIE1,TXIE         ;Is the transmitter interrupt enabled?
            return                  ;No, done
            
            BANK0                   ;Switch to bank-0
            btfss PIR1,TXIF         ;Yes. Is there an transmit interrupt
            return                  ;No, Async TX services not required

            ;--- Transmit interrupt services required
            BANK1                   ;Switch to bank-1
            movf  tx_comm_pntr,w    ;Get current TX buffer position
            incf  tx_comm_pntr,f    ;Post-Inc pointer for next time
            movwf FSR               ;Set indirect pointer
            movf  INDF,w            ;Read the next outgoing byte            
            BANK0                   ;Switch to bank-0
            movwf TXREG             ;Feed the transmit register
            
            BANK1                   ;Switch to bank-1
            decfsz tx_cntr,f        ;Dec and update the tx count
            return                  ;Not zero, more characters to go
            
            ;--- Last character is being sent
            bsf comm_aux_reg,TX_LAST_CHAR_BIT   ;Indicate last character is being sent
            bcf PIE1,TXIE                       ;Disable future transmitter interrupts 

            return                  ;done
            
             
;********************************************************************** 
; CYCLE_SRV: Interrupt service routine support for the periodic interrupts 
; generated when TIMER-0 rolls over.

cycle_srv:    ;--- Check for periodic interrupts
              btfss INTCON, T0IF      ;Test timer0 overflow flag
              return                  ;Not a timer0 interrupt
            
              ;--- Process sub-cycle (1.64ms) interrupts
              bcf INTCON, T0IF        ;Clear timer0 overflow flag
              
              ;--- Do TX last character RTX/TXEN function
              BANK1                                 ;Switch to bank-1
              btfss comm_ctrl_reg,TX_BUSY_BIT       ;Are transmissions active?
              goto cyc0                             ;No, skip around
              
              btfss comm_aux_reg,TX_LAST_CHAR_BIT   ;Yes, but on last character?
              goto cyc0                             ;No, skip around
             
              btfss TXSTA,TRMT                      ;Has last character finished?  
              goto cyc0                             ;No, skip around
            
              bcf comm_aux_reg,TX_LAST_CHAR_BIT     ;Indicate no longer on last character
              bcf comm_ctrl_reg,TX_BUSY_BIT         ;Indicate transmitter is now free
       
              BANK0                                 ;Switch to bank-0
              bsf PORTC,RTS_BIT                     ;Deactivate RTS / Disable RS485 driver
cyc0:              
             
              ;************************************
              ;*** Add user sub-cycle code here
              ;*** Code will execute every 1.64ms
              ;************************************
              
              BANK0                   ;Switch to bank-0 
              decfsz timer0_sub_cyc,f ;Dec and zero-test the sub-cycle counter  
              return                  ;Not zero, all done
              
              ;--- Process full-cycle (13.1ms) interrupts
              movlw SUB_CYC_MODULUS   ;Get sub-cycle preload value
              movwf timer0_sub_cyc    ;reload the sub-cycle counter
            
              bsf opsys_var0, CYC_SYNC_BIT  ;Inform upper level its time for action
            
              ;************************************
              ;*** Add user full-cycle code here
              ;*** Code will execute every 13.1ms
              ;************************************

              return
            

;********************************************************************** 
; PSP_SRV: Interrupt service routine provides support for interrupts
; when the host performs a read (IOR) or write (IOW) operation to the
; PIC16C74 Parallel-Slave-Port.  

psp_srv:    ;--- Check for PSP interrupts
            BANK0                 ;Switch to bank 0
            btfss PIR1,PSPIF      ;Test for PSP interrupt
            return
              
            ;--- Support for PSP interrupt routines
            bcf PIR1,PSPIF        ;Clear PSP interrupt
            
            BANK1                 ;Switch to bank 1
            btfsc TRISE,IBOV      ;Overflow bit set?
            bcf TRISE,IBOV        ;Yes, clear it
            
            BANK0                 ;Switch to bank 0
            btfss PORTA,4         ;Check A0, Data or Command ?
            goto psp_data         ;Addr = 0, means its data
           
            ;--- Perform PSP commands
            BANK1                 ;Switch to bank 1
            btfsc TRISE,IBF       ;Did host write a command?
            goto psp_wr_cmd       ;Yes, branch
            
            ;--- Command reads not supported, cleanup and exit
            return

            ;--- A command was written, determine which one to perform
psp_wr_cmd: BANK0                 ;Switch to bank 0
            
            ;--- CMD0 - Set Auto-Inc mode and data address pointer
            movf PORTD,w          ;Read PORTD input latch (clears IBF flag)
            andlw b'11000000'     ;Test if bit 7=0 & 6=0
            btfss STATUS,Z        ;Command 0?
            goto cmd1             ;No, branch
            
            bsf psp_ctrl_reg, PSP_AUTO_INC_BIT   ;Assume auto-inc is desired
            btfss PORTD, PSP_AUTO_INC_BIT        ;Was Assumption correct?
            bcf psp_ctrl_reg, PSP_AUTO_INC_BIT   ;No, clear bit
             
            movf PORTD,w              ;Read PSP input latch
            andlw PSP_ARY_SIZE-1      ;Extract just address bits
            movwf psp_data_ary_pntr   ;Store new index
            
            addlw psp_read_ary    ;Calc offset into output data array
            movwf FSR             ;Set indirect pointer
            movf INDF,w           ;Read the value out of the array
            movwf PORTD           ;Overwrite PSP output latch with new data
         
            return
             
cmd1:       ;--- CMD1 - Unassigned
            movf PORTD,w          ;Read PORTD input latch  
						andlw b'11000000'			;Keep only command bits
						xorlw b'01000000'     ;Test if bit 7=0 & 6=1    
            btfss STATUS,Z        ;Command 1?
            goto cmd2             ;No, branch
            ;************************
            ;*** Add user code here
						;************************
            return

cmd2:       ;--- CMD2 - Unassigned
            movf PORTD,w          ;Read PORTD input latch 
						andlw b'11000000'			;Keep only command bits
						xorlw b'10000000'     ;Test if bit 7=1 & 6=0
            btfss STATUS,Z        ;Command 2?
            goto cmd3             ;No, branch
            ;************************
            ;*** Add user code here
            ;************************
            return

cmd3:       ;--- CMD3 - Clear program interupts
            ;**********************************************
            ;*** Add user code here to clear program interupts 
            ;**********************************************
            movf PORTD,w          ;Read PORTD input latch (clears IBF flag)
            return
            
psp_data:   ;--- Do PSP data transfers
            BANK1                 ;Switch to bank 1
            btfsc TRISE,IBF       ;Did host write data?
            goto psp_wr_data      ;Yes, branch

psp_rd_data:
            ;--- Read data out of IOCP-74 data output array 
            BANK0                 ;Switch to bank 0

            btfsc psp_ctrl_reg,PSP_AUTO_INC_BIT   ;Auto-Inc enabled?
            incf psp_data_ary_pntr,f              ;Yes, pre-inc the pointer
            
            movf psp_data_ary_pntr,w    ;Read the pointer
            andlw PSP_ARY_SIZE-1        ;Keep within range of array

            addlw psp_read_ary    ;Calc offset into output data array
            movwf FSR             ;Set indirect pointer
            movf INDF,w           ;Read a value out of the array
            movwf PORTD           ;Update PSP output latch for next read
           
            return

psp_wr_data:
            ;--- Write data into IOCP-74 input array
            BANK0                       ;Switch to bank 0
            
            movf psp_data_ary_pntr,w    ;Read the pointer
            andlw PSP_ARY_SIZE-1        ;Keep within range of array
						movwf	intr_temp0						;Save a copy of it for semaphore calcs

            addlw psp_write_ary   ;Calc offset into input data array
            movwf FSR             ;Set indirect pointer
            
            movf PORTD,w          ;Read PORTD input latch (clears IBF flag)
            movwf INDF            ;Store the data into the array

         		movlw HIGH get_semaphore  		;Get high byte of table address         
         		movwf PCLATH          				;Make sure program counter is set properly

  					movf	intr_temp0,w						;Get relative array address
						andlw	b'00000111'							;Determine corresponding bit (0-7)

						call get_semaphore						;Look up the bit mask
						movwf intr_temp1							;Store it away

						movlw b'11111000'							;Load w with byte number mask
						andwf intr_temp0,f						;Keep only byte number bits			
  					rrf intr_temp0,f							;Convert to byte offset
						rrf intr_temp0,f
						rrf intr_temp0,f
						movlw	psp_wr_flags+0					;W = starting address of wr flags
						addwf	intr_temp0,w						;Add in offset

						movwf	FSR											;Put in indirect pointer
						movf	intr_temp1,w						;Recall semaphore bit mask
						iorwf	INDF,f									;Set the wr flag to indicate new data from host

            btfsc psp_ctrl_reg,PSP_AUTO_INC_BIT   ;Auto-Inc enabled?
            incf psp_data_ary_pntr,f              ;Yes, post-inc the pointer
           
            return

;**************************************************************************
; GET_SEMAPHORE -- Determines bit mask for corresponding table offset
;
;	Input			:	Enter with w containing relative byte offset
;
;	Outoput		:	w returns with bit mask (a single bit set) 
;	
;	Uses			:	w

get_semaphore: 	;--- Table of semaphore flag masks
					  		addwf PCL,f   ;Add offset to pc, figure table entry
            		retlw b'00000001'     ;Bit 0  
            		retlw b'00000010'     ;Bit 1
            		retlw b'00000100'     ;Bit 2
            		retlw b'00001000'     ;Bit 3
            		retlw b'00010000'     ;Bit 4
            		retlw b'00100000'     ;Bit 5
								retlw b'01000000'     ;Bit 6
								retlw b'10000000'     ;Bit 7

;************************************************************           
            end   ;End of IOCPbios
